//
// srcrt.cc       : Source Route Filter
// author         : Fabio Silva
//
// Copyright (C) 2000-2001 by the Unversity of Southern California
// $Id: srcrt.cc,v 1.1 2001/09/07 23:11:49 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include "srcrt.hh"

SrcRt *app;

void MyFilterReceive::recv(Message *msg, handle h)
{
  app->recv(msg, h);
}

void SrcRt::recv(Message *msg, handle h)
{
  Message *return_msg = NULL;

  if (h != filterHandle){
    fprintf(stderr, "Error: Received a message for handle %d when subscribing to handle %d !\n", h, filterHandle);
    return;
  }

  return_msg = ProcessMessage(msg);

  if (return_msg)
    ((DiffusionRouting *)dr)->sendMessageToNext(msg, h);
}

Message * SrcRt::ProcessMessage(Message *msg)
{
  char *original_route, *new_route, *p;
  int len, total_len;
  int32_t next_hop;
  NRSimpleAttribute<char *> *route = NULL;

  route = SourceRouteAttr.find(msg->msg_attr_vec);
  if (!route){
    fprintf(stderr, "Error: Can't find the route attribute !\n");
    return msg;
  }

  original_route = route->getVal();
  len = strlen(original_route);

  // Check if we are the last hop
  if (len == 0)
    return msg;

  // Get the next hop
  next_hop = atoi(original_route);

  // Remove last hop from source route
  p = strstr(original_route, ":");
  if (!p){
    // There's just one more hop
    new_route = new char[1];
    new_route[0] = '\0';
  }
  else{
    p++;
    len = strlen(p);
    new_route = new char[(len + 1)];
    strncpy(new_route, p, (len + 1));
    if (new_route[len] != '\0')
      fprintf(stderr, "Warning: String must end with NULL !\n");
  }

  route->setVal(new_route);

  // Free memory
  delete [] new_route;

  // Send the packet to the next hop
  msg->next_hop = next_hop;
  ((DiffusionRouting *)dr)->sendMessage(msg, filterHandle);

  delete msg;

  return NULL;
}

handle SrcRt::setupFilter()
{
  NRAttrVec attrs;
  handle h;

  // Match all packets with a SourceRoute Attribute
  attrs.push_back(SourceRouteAttr.make(NRAttribute::EQ_ANY, ""));

  h = ((DiffusionRouting *)dr)->addFilter(&attrs, SRCRT_FILTER_PRIORITY, fcb);

  ClearAttrs(&attrs);
  return h;
}

void SrcRt::run()
{
  // Doesn't do anything
  while (1){
    sleep(100);
  }
}

void SrcRt::usage()
{
  fprintf(stderr, "Usage: srcrt [-p port] [-h]\n\n");
  fprintf(stderr, "\t-p - Uses port 'port' to talk to diffusion\n");
  fprintf(stderr, "\t-h - Prints this information\n");
  fprintf(stderr, "\n");
  exit(0);
}

SrcRt::SrcRt(int argc, char **argv)
{
  int opt;
  u_int16_t diffusion_port = DEFAULT_DIFFUSION_PORT;

  // Parse command line options
  opterr = 0;

  while (1){
    opt = getopt(argc, argv, "hp:");
    switch (opt){

    case 'p':

      diffusion_port = (u_int16_t) atoi(optarg);
      if ((diffusion_port < 1024) || (diffusion_port >= 65535)){
	fprintf(stderr, "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage();

      break;

    case '?':

      fprintf(stderr, "Error: %c isn't a valid option or its parameter is missing !\n", optopt);
      usage();

      break;

    case ':':

      fprintf(stderr, "Parameter missing !\n");
      usage();

      break;

    }

    if (opt == -1)
      break;
  }

  // Create Diffusion Routing class
  dr = NR::createNR(diffusion_port);
  fcb = new MyFilterReceive;

  // Set up the filter
  filterHandle = setupFilter();
  fprintf(stderr, "SrcRt filter received handle %d\n", filterHandle);
  fprintf(stderr, "SrcRt filter initialized !\n");
}

int main(int argc, char **argv)
{
  // Initialize and run the Source Route Filter
  app = new SrcRt(argc, argv);
  app->run();

  return 0;
}
