/* VERSION WINSNG2-1_0 	 Jun 29 2001  */
/*
	rf.c - API routines for the WINSNG 2.0 RF Modem
    Copyright (C) 2000 Sensoria Corporation    
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 	Revision history:
    
*/    
#include "rf.h"
#include <stdio.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>

/*Internally visible function prototypes*/
int IsHexDigit(char c);

/*Allocate initial storage for a packet
 *Specify size in terms of data size, do not include header size
 *Return -1 on failure, 1 on success
 */
int initializeRfPacket(RfPacket *p,size_t size)
{
  /*Must leave enough room for a header*/
  size += sizeof(RfHeader);
    
  p->bytesAlloc = size;
  p->raw = (uint8 *) malloc(size);
  if (p->raw == NULL)
  {
    fprintf(stderr,"initializePacket(): Failed to allocate memory");
    return -1;
  }
  p->header = (RfHeader *) p->raw;
  p->data   = p->raw + sizeof(RfHeader);
  return 1;
}

/*Create RF Packet from header and data*/
int createRfPacket(RfPacket *p,RfHeader *header,uint8 *data, size_t size)
{
  /*Allocate storage for packet if new size is greater than old size
   *First delete storage if it was previously allocated (via realloc)
   */
  int newSize = size + sizeof(RfHeader);
  if (newSize > p->bytesAlloc) 
  {
    realloc( p->raw, newSize); 
    if (p->raw == NULL)
    {
      fprintf(stderr,"createRfPacket(): Failed to allocate memory.\n");
      return -1;
    }
    p->bytesAlloc = newSize;
  }
 
  /*Copy header to packet*/
  memcpy(p->raw, header, sizeof(RfHeader) );
  /*Copy data to packet*/
  memcpy(p->raw + sizeof(RfHeader), data, size );  
  
  p->header = (RfHeader *) p->raw;
  p->data   = p->raw + sizeof(RfHeader);
  return 1;  
}

/*Read a packet from driver
 *Return -1 on error, Bytes read on success	(including header)
 *Will automatically deallocate previous data stored in argument packet
 */
int readPacket(int fd,RfPacket *packet)
{
  int bytesRead;
  fd_set rfds;  
  int retval;
  int totalBytes;
  
  /*Check to see if a packet is already available*/
  totalBytes=nextPacketLength(fd);
  //fprintf(stderr,"Driver reports totalbytes=%d\n",totalBytes);
  if(totalBytes == -1)
    return -1;
    
  /*If there is no available packet, wait for one*/
  if (totalBytes == 0)
  { 
    /*Setup select structures*/
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);  
    /*Wait for packet to arrive*/  
    retval=select(fd+1, &rfds, NULL,NULL,NULL);
     
    if (retval == -1)
      return -1;
    /*Determine size of next packet*/
    totalBytes=nextPacketLength(fd);    
  }
  
  /*Allocate storage for packet if new size is greater than old size
   *First delete storage if it was previously allocated (via realloc)
   */
  if (totalBytes > packet->bytesAlloc) 
  {
    packet->raw = realloc( packet->raw, totalBytes); 
    if (packet->raw == NULL)
    {
      fprintf(stderr,"readPacket(): Failed to allocate memory.\n");
      packet->bytesAlloc=0;
      return -1;
    }
    packet->bytesAlloc=totalBytes;
  }

  /*Read in entire packet*/
  /*Driver is "atomic" in that one read will return one packet, 
    regardless of how many bytes it was asked to read, as long
    as the value given is large enough to hold a packet*/
  bytesRead=read(fd,packet->raw,totalBytes);  
  if (bytesRead < 0)
  {
    return -1;
  }    
  
  packet->header = (RfHeader *) packet->raw;
  packet->data   = packet->raw + sizeof(RfHeader);
  return bytesRead;  
}

/*Send a packet
 *Return -1 on failure, number of bytes sent on success
 *The header and data pointers of the argument RfPacket must
 *be contiguous in memory. The function readPacket guarantees
 *this, but if it is not used this criterion must still be met.
 */

int sendPacket(int fd,RfPacket *packet)
{
  int bytesToWrite;
  
  /*Number of bytes to write is size of header as well as any data*/
  bytesToWrite=sizeof(RfHeader) + packet->header->length;
  return(write(fd,packet->raw,bytesToWrite) );
}

/*Determine if driver is connected to a base or remote*/
/*Return 0  if a driver is connected to a remote*/
/*Return 1  if a driver is connected to a base*/
/*Return -1	on error*/
int determineBaseRemote(int fd)
{
  int status;
  char ToDriver[4];
  
  status=ioctl(fd ,	Determine_Base_Remote_Ctrl,ToDriver);
  if (status==-1)
    return -1;
  
  return (int) ToDriver[0];                    
}

/*Set an RFModem as base*/
/*Return 0 on success, -1 on failure*/
int setBase(int fd)
{
  int status;
  char ToDriver[4];
  
  status=ioctl(fd ,Set_Base_Ctrl ,ToDriver);
  if (status==-1)
    return -1;
  
  return 0;                    
}

/*Set an RFModem as base*/
/*Return 0 on success, -1 on failure*/
int setRemote(int fd)
{
  int status;
  char ToDriver[4];
  
  status=ioctl(fd ,Set_Remote_Ctrl ,ToDriver);
  if (status==-1)
    return -1;
  
  return 0;                    
}

/*Determine the network of an RFModem
  Returns a number between 0 and 63 on success
  Returns -1 on failure
*/
int getNetwork(int fd)
{
  int  status;
  char ToDriver[4];
  
  status=ioctl(fd ,Determine_Network_Ctrl ,ToDriver);
  if (status==-1)
    return -1;
  
  return( ntohl( * ((int *) ToDriver) )  );
}

/*Set network number.
  Must provide a network number between 0 and 63
  Returns -1 on failure
*/
int setNetwork(int fd,int network)
{
  int  status;
  char ToDriver[4];
  
  if (network < 0 || network > 63)
    return -1;
  
  status=ioctl(fd ,Set_Network_Ctrl ,ToDriver);
  if (status==-1)
    return -1;
  
  return 0;  
}

/*Store address in a
 *Return -1 on failure, 1 on success*/
int getAddress(int fd,Address a)
{
  int status;
  unsigned char ToDriver[4];
  
  status= ioctl(fd ,	Determine_Address_Ctrl, ToDriver);
  if (status==-1)
    return -1;
  copyAddress(a,ToDriver);
  return 1;  
}

/*Return RSSI value
  Valid values are 0-255
  Returns -1 on error
*/
int getRSSI(int fd)
{
  int result;
  char ToDriver[4];
  
  result=ioctl(fd, Determine_Signal_Strength_Ctrl, ToDriver);
  if (result!=0)
  {
    return -1;
  }
  return( ntohl( * ((int *) ToDriver) )  );
}

/*Set the transmit power of the RF Modem
  0 = 10  mW
  1 = 100 mW
  Return -1 on failure, 0 on success
*/
int setTransmitPower(int fd, int Power)
{
  int result;
  char ToDriver[4];
  
  if ( (Power!=0) && (Power!=1) ) 
    return -1;
  
  ToDriver[0] = (char) Power;  
  result=ioctl(fd, Set_Transmit_Power_Ctrl , ToDriver);
  if (result!=0)
  {
    return -1;
  }

  return 0;
}

/*Determine the transmit power of the RF Modem
  Returns
  0 = 10  mW
  1 = 100 mW
  -1 on failure
*/
int getTransmitPower(int fd)
{
  int result;
  char ToDriver[4];
  
  result=ioctl(fd, Determine_Transmit_Power_Ctrl, ToDriver);
  if (result!=0)
  {
    return -1;
  }
  return( ntohl( * ((int *) ToDriver) )  );
}

/*Turn on or off connect/disconnect notifications from the driver.
  Return 0 on success, -1 on failure
*/
int setNotifications(int fd, int setting)
{
  int result;
  char ToDriver[4];
  
  if ( (setting!=0) && (setting!=1) ) 
    return -1;
  
  ToDriver[0] = (char) setting;  
  result=ioctl(fd, Set_Notifications , ToDriver);
  if (result!=0)
  {
    return -1;
  }

  return 0;
}

/*Return next packet length.
  0 if non available
  -1 on error
*/
int nextPacketLength(int fd)
{
  int result;
  char toDriver[4];
  
  result=ioctl(fd, Ctrl_QueryFrontSize, toDriver);
  if (result!=0)
  {
    return -1;
  }
  return ( ntohl( * ((int *) toDriver) ) );
}

/*Get connected RF Modem List
 *fd  =Device file of RF Modem driver
 *list=Pointer to where list will be stored
 *max =maximum number of connected addresses to store
 *Returns number of connected addresses found (each 3 bytes)
 * or -1 on failure*/
int getConnectedList(int fd,uint8* list,int max)
{
  unsigned char ToDriver[MaxListBytes + 1];
  int  num;
  int  status;
  
  /*Do not allow more than 20 modems*/
  if (max > MaxListBytes/3)
    max=MaxListBytes/3;
    
  /*Ask driver for connected list*/
  status=ioctl(fd,Get_Connected_List,ToDriver);
  if(status == -1)
    return -1; 

  /*Determine how big the list is
    Driver stores this in first byte
  */ 
  num=(int) ToDriver[0];
  if (num > max) 
    num=max;  

  /*Copy addreses to caller's address space*/                   
  memcpy(list,ToDriver + 1, num * 3);
  
  return num;
  
}


/*Copy src into dest
  Caution- caller must ensure src and dest point to 3 valid bytes*/
void copyAddress(Address dest,Address src)
{
  dest[0]=src[0];
  dest[1]=src[1];
  dest[2]=src[2];  
}

/*Put RF Modem to sleep
  Return -1 on error
*/
int sleepRfModem(int fd)
{
  int result;
  char toDriver[4];
  
  result=ioctl(fd, RF_Sleep_Ctrl, toDriver);
  if (result!=0)
  {
    return -1;
  }
  return 1;
}

/*Wake sleeping RF Modem 
  Return -1 on error
*/
int wakeRfModem(int fd)
{
  int result;
  char toDriver[4];
  
  result=ioctl(fd, RF_Wake_Ctrl, toDriver);
  if (result!=0)
  {
    return -1;
  }
  return 1;
}

/*Return zero if two addresses don't match, nonzero otherwise*/
int compareAddress(Address a1,Address a2)
{
  return ( a1[0]==a2[0] && 
           a1[1]==a2[1] &&
           a1[2]==a2[2]
         );
}

/*  Return 1 if read address was valid, and store it into address
    Return 0 if read address was invalid
    Caller must ensure buff points to at least 8 bytes */
int sreadAddress(char* buff,Address address)
{
	 int value;
  
  /*Ensure we have periods in the proper positions*/
  if (buff[2]!='.' || buff[5]!='.')
    return 0;
   
  /*Ensure all characters are hex digits*/
  if (!(IsHexDigit(buff[0]) && IsHexDigit(buff[1]) &&
        IsHexDigit(buff[3]) && IsHexDigit(buff[4]) &&
        IsHexDigit(buff[6]) && IsHexDigit(buff[7])
     ) )
    return 0;
    
  /*Response is valid. Parse and copy*/
  sscanf(&buff[0], "%02x", &value); address[0]=(uint8) value;  
  sscanf(&buff[3], "%02x", &value); address[1]=(uint8) value;
  sscanf(&buff[6], "%02x", &value); address[2]=(uint8) value;  
  return 1;  
}

/*Read in a 24 bit comma separated hex address
  Return 1 if read address was valid, and store it into address
  Return 0 if read address was invalid*/
int readAddress(FILE* fp,Address address)
{   			   
  char buff[9]; /*  Space for	XX.XX.XX/0 */

  /*Read in string from user*/
  fscanf(fp,"%8s",buff);  
  return sreadAddress(buff,address);
}

/*Print out an address in a format XX.XX.XX*/
void printAddress(FILE* fp,Address address)
{
  fprintf(fp,"%02X.%02X.%02X",address[0],address[1],address[2]);  
}

void sprintAddress(char* c,Address address)
{
  sprintf(c,"%02X.%02X.%02X",address[0],address[1],address[2]);  
}

/*Return 1 if the character is in the range
  0-9 or
  a-f	or
  A-F
  Return 0 otherwise*/
int IsHexDigit(char c)
{
  return( (c>='0' && c<='9') ||
          (c>='a' && c<='f') ||
          (c>='A' && c<='F') );
}


